/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include <ctype.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#include <string.h>

#include "http.h"

char *extension_mime_types[40][2]={
  { ".html", "text/html" },
  { ".htm",  "text/html" },
  { ".shtml","text/html" },
  { ".css",  "text/css" },
  { ".txt",  "text/plain" },
  { ".mdb",  "application/msaccess" },
  { ".xls",  "application/msexcel" },
  { ".doc",  "application/msword" },
  { ".exe",  "application/octet-stream" },
  { ".pdf",  "application/pdf" },
  { ".rtf",  "application/rtf" },
  { ".tgz",  "application/x-compressed" },
  { ".gz",   "application/x-compressed" },
  { ".z",    "application/x-compress" },
  { ".swf",  "application/x-shockwave-flash" },
  { ".tar",  "application/x-tar" },
  { ".rar",  "application/x-rar-compressed" },
  { ".torrent",  "application/x-bittorrent" },
  { ".rss",  "application/x-alluvium" },
  { ".zip",  "application/x-zip-compressed" },
  { ".ra",   "audio/x-pn-realaudio" },
  { ".ram",  "audio/x-pn-realaudio" },
  { ".wav",  "audio/x-wav" },
  { ".gif",  "image/gif" },
  { ".jpeg", "image/jpeg" },
  { ".jpe",  "image/jpeg" },
  { ".jpg",  "image/jpeg" },
  { ".png",  "image/png" },
  { ".avi",  "video/avi" },
  { ".ogg",  "audio/x-ogg" },
//    { ".ogg",  "audio/ogg" },
//    { ".ogg",  "application/x-ogg" },
//    { ".ogg",  "audio/x-vorbis" },
  { ".mp3",  "audio/mpeg" },
  { ".mpeg", "video/mpeg" },
  { ".mpg",  "video/mpeg" },
  { ".qt",   "video/quicktime" },
  { ".mov",  "video/quicktime" },
  { NULL, NULL }
};


char *guess_mime_type(char *filename)
{
  int i;
  char *extension;

  extension = strrchr(filename, '.');
  if (extension == NULL) {
    return "text/plain";
  }

  for (i = 0;  extension_mime_types[i][0] != NULL; i++) {
    if (strcmp(extension, extension_mime_types[i][0]) == 0) {
      return extension_mime_types[i][1];
    }
  }

  return "application/octet-stream";
}


char *str_trim_in_place(char *buffer) {
  int n;

  // remove any white space at the end of the line
  n = strlen(buffer) - 1;
  while (n >= 0 && isspace(buffer[n])) {
    buffer[n] = '\0';
    n--;
  }

  n = 0;
  while (n < strlen(buffer) && isspace(buffer[n])) {
    n++;
  }

  return &buffer[n];
}


void http_parse_request(char *headers, http_request *request)
{
  int offset = 0;
  int i;

  headers = str_trim_in_place(&headers[offset]);

  if (strncasecmp(&headers[offset], "get ", 4) == 0) {
    request->type = HTTP_GET;

    offset += 4;

    while (isspace(headers[offset])) {
      offset++;
    }

    i = 0;
    while ((offset < sizeof(request->filename)) &&
           ((headers[offset] != '\r') &&
            (headers[offset] != '\n') &&
            (headers[offset] != ' '))) {
      request->filename[i] = headers[offset];
      i++;
      offset++;
    }
    request->filename[i] = 0;

    headers = str_trim_in_place(&headers[offset]);

    if (strlen(headers) == 0) {
      request->version = 0.8; //?

      if (strncasecmp(request->filename, "http://", 7)) {
        // strip "http://", then strip the server name
        char *foo = strchr(&request->filename[7], '/');
        if (foo != NULL) {
          char *buff = (char*)malloc(sizeof(request->filename));
          snprintf(buff, sizeof(buff), "%s", foo);
          snprintf(request->filename, sizeof(request->filename), "%s", buff);
          free(buff);
        }
      }
    } else {
      if (strncasecmp(headers, "HTTP/", 5) == 0) {
        request->version=atof(&headers[5]);
      }
    }
    //printf("Request Version: %.1f\n",request->version);
  } else if (strncasecmp(&headers[offset], "head ", 5) == 0) {
    request->type = HTTP_HEAD;

    offset += 5;

    while (isspace(headers[offset])) {
      offset++;
    }

    i = 0;
    while ((offset < sizeof(request->filename)) &&
           ((headers[offset] != '\r') &&
            (headers[offset] != '\n') &&
            (headers[offset] != ' '))) {
      request->filename[i] = headers[offset];
      i++;
      offset++;
    }
    request->filename[i] = 0;

    headers = str_trim_in_place(&headers[offset]);

    if (strlen(headers) == 0) {
      request->version = 0.8; //?

      if (strncasecmp(request->filename, "http://", 7)) {
        // strip "http://", then strip the server name
        char *foo = strchr(&request->filename[7], '/');
        if (foo != NULL) {
          char *buff = (char*)malloc(sizeof(request->filename));
          snprintf(buff, sizeof(buff), "%s", foo);
          snprintf(request->filename, sizeof(request->filename), "%s", buff);
          free(buff);
        }
      }
    } else {
      if (strncasecmp(headers, "HTTP/", 5) == 0) {
        request->version=atof(&headers[5]);
      }
    }
    //printf("Request Version: %.1f\n",request->version);
  } else if (strncasecmp(&headers[offset], "post ", 5) == 0) {
    request->type = HTTP_POST;
  } else if (strncasecmp(&headers[offset], "http/1.0 200 ok", 15) == 0) {
    request->type = HTTP_OK;
  } else {
    request->type = HTTP_UNKNOWN;
  }
}


// this is a cheesy little function to do a sort of okay job of
// parsing HTTP headers from a buffer.
void http_parse_headers(char *headers, http_request *request)
{
  char *key, *value;
  int offset = 0;
  int len = strlen(headers);

  request->num_headers = 0;

  while ((request->num_headers < MAX_HTTP_HEADERS) && (offset < len)) {
    // parse out the key part of the header
    key = &headers[offset];

    while ((offset < len) &&
           (headers[offset] != ':')) {
      offset++;
    }
    offset++;

    if (offset >= len) {
      break;
    }

    headers[offset - 1] = 0;
    request->keys[request->num_headers] = str_trim_in_place(key);

    // parse out the value part of the header
    value = &headers[offset];

    while ((offset < len) &&
           (headers[offset] != 0x0D) &&
           (headers[offset] != 0x0A)) {
      offset++;
    }

    while ((offset < len) &&
           ((headers[offset] == 0x0D) ||
            (headers[offset] == 0x0A))) {
      headers[offset] = 0;
      offset++;
    }

    request->values[request->num_headers] = str_trim_in_place(value);

    request->num_headers++;
  }

  return;
}


// searches an iobuf for something that looks like the end of some
// HTTP headers, and returns the offset after the end of the headers
// if it is found.
int http_find_header_end(io_buffer *iobuf, int end_header)
{
  int offset;
  int bytes_avail;

  bytes_avail = io_buffer_bytes_used(iobuf);

  offset = 0;
  while (offset < bytes_avail) {
    if (end_header) {
      if ((offset <= bytes_avail - 4) &&
          (HTTP_END_HEADERS == io_buffer_peek_int(iobuf, 4, offset))) {
        return offset + 4;
      }
    } else {
      if ((offset <= bytes_avail - 2) &&
          (HTTP_END_HEADER == io_buffer_peek_int(iobuf, 2, offset))) {
        return offset + 2;
      }
    }

    offset++;
  }

  return -1;
}


int hex2int(char *hexstr)
{
  int high;
  int low;
  int result;

  high = hexstr[0];
  if ('0' <= high && high <= '9') {
    high -= '0';
  } else if ('a' <= high && high <= 'f') {
    high -= ('a' - 10);
  } else if ('A' <= high && high <= 'F') {
    high -= ('A' - 10);
  }

  low = hexstr[1];
  if ('0' <= low && low <= '9') {
    low -= '0';
  } else if ('a' <= low && low <= 'f') {
    low -= ('a' - 10);
  } else if ('A' <= low && low <= 'F') {
    low -= ('A' - 10);
  }

  result = (high * 16) + low;
  return result;
}


void swapchar(char *string, char oldchar, char newchar)
{
  while (*string != 0) {
    if (*string == oldchar) {
      *string = newchar;
    }

    string++;
  }
}


void decode_url(char *encoded)
{
  char *decoded = NULL;

  decoded = encoded;
  swapchar(decoded, '+', ' ');

  while (*encoded != 0) {
    if (*encoded == '%') {
      encoded++;

      if (isxdigit(encoded[0]) &&
          isxdigit(encoded[1])) {
        *decoded = (char)hex2int(encoded);

        decoded++;
        encoded += 2;
      }
    } else {
      *decoded = *encoded;

      encoded++;
      decoded++;
    }
  }
}


void fix_slashes(char *original)
{
#ifdef WIN32
  swapchar(original, '/', '\\');
#else
  swapchar(original, '\\', '/');
#endif
}
